/**
 * Ext GWT 3.0.0-beta1 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.widget.core.client;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import com.google.gwt.cell.client.Cell;
import com.google.gwt.cell.client.Cell.Context;
import com.google.gwt.cell.client.ValueUpdater;
import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.EventTarget;
import com.google.gwt.dom.client.Node;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.resources.client.ClientBundle;
import com.google.gwt.resources.client.CssResource;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.resources.client.ImageResource.ImageOptions;
import com.google.gwt.resources.client.ImageResource.RepeatStyle;
import com.google.gwt.safehtml.shared.SafeHtml;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.safehtml.shared.SafeHtmlUtils;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Event;
import com.sencha.gxt.core.client.GXT;
import com.sencha.gxt.core.client.ValueProvider;
import com.sencha.gxt.core.client.dom.CompositeElement;
import com.sencha.gxt.core.client.dom.DomHelper;
import com.sencha.gxt.core.client.dom.XDOM;
import com.sencha.gxt.core.client.dom.XElement;
import com.sencha.gxt.core.client.util.Util;
import com.sencha.gxt.data.shared.ListStore;
import com.sencha.gxt.data.shared.event.StoreAddEvent;
import com.sencha.gxt.data.shared.event.StoreClearEvent;
import com.sencha.gxt.data.shared.event.StoreDataChangeEvent;
import com.sencha.gxt.data.shared.event.StoreFilterEvent;
import com.sencha.gxt.data.shared.event.StoreHandlers;
import com.sencha.gxt.data.shared.event.StoreRecordChangeEvent;
import com.sencha.gxt.data.shared.event.StoreRemoveEvent;
import com.sencha.gxt.data.shared.event.StoreSortEvent;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent;
import com.sencha.gxt.widget.core.client.cell.DefaultHandlerManagerContext;
import com.sencha.gxt.widget.core.client.event.RefreshEvent;
import com.sencha.gxt.widget.core.client.event.RefreshEvent.HasRefreshHandlers;
import com.sencha.gxt.widget.core.client.event.RefreshEvent.RefreshHandler;
import com.sencha.gxt.widget.core.client.event.SelectEvent;
import com.sencha.gxt.widget.core.client.event.XEvent;
import com.sencha.gxt.widget.core.client.tips.QuickTip;

/**
 * Displays a list of data.
 * 
 * @param <M> the model type
 * @param <N> the cell data type
 */
public class ListView<M, N> extends Component implements HasRefreshHandlers {

  public interface ListViewAppearance<M> {

    Element findCellParent(XElement item);

    Element findElement(XElement child);

    List<Element> findElements(XElement parent);

    void onOver(XElement item, boolean over);

    void onSelect(XElement item, boolean select);

    /**
     * Renders the container.
     * 
     * @param builder the builder
     */
    void render(SafeHtmlBuilder builder);

    /**
     * Optionally renders extra markup at the end of the the list.
     * 
     * @param builder the builder
     */
    void renderEnd(SafeHtmlBuilder builder);

    /**
     * Renders a single item.
     * 
     * @param builder the builder
     * @param content the item content
     */
    void renderItem(SafeHtmlBuilder builder, SafeHtml content);

  }

  public interface ListViewDefaultResources extends ClientBundle {
    @Source("ListView.css")
    ListViewDefaultStyle css();

    @ImageOptions(repeatStyle = RepeatStyle.Horizontal)
    ImageResource selectedBackground();
  }

  public interface ListViewDefaultStyle extends CssResource {

    String item();

    String over();

    String sel();

    String view();
  }

  private static ListViewDefaultResources standardResources;

  public static ListViewDefaultResources getStandardResources() {
    if (standardResources == null) {
      standardResources = GWT.create(ListViewDefaultResources.class);
      standardResources.css().ensureInjected();
    }
    return standardResources;
  }

  protected ListViewAppearance<M> appearance;
  protected int rowSelectorDepth = 5;
  protected ListStore<M> store;

  final protected ValueProvider<M, N> valueProvider;

  private CompositeElement all;
  private Cell<N> cell;
  private boolean enableQuickTip = true;

  private HandlerRegistration handlerRegistration;
  private String loadingText;
  private XElement overElement;
  private QuickTip quickTip;
  private boolean selectOnHover;
  private ListViewSelectionModel<M> sm;
  private StoreHandlers<M> storeHandlers;

  /**
   * Creates a new list view.
   * 
   * @param store the store
   * @param valueProvider the value provider
   */
  public ListView(ListStore<M> store, ValueProvider<M, N> valueProvider) {
    this(store, valueProvider, new ListViewDefaultAppearance<M>());
  }

  /**
   * Creates a new list view.
   * 
   * @param store the store
   * @param valueProvider the value provider
   * @param cell the cell
   */
  public ListView(ListStore<M> store, ValueProvider<M, N> valueProvider, Cell<N> cell) {
    this(store, valueProvider, new ListViewDefaultAppearance<M>());
    setCell(cell);
  }

  /**
   * Creates a new list view.
   * 
   * @param store the store
   * @param valueProvider the value provider
   * @param appearance the appearance
   */
  public ListView(final ListStore<M> store, ValueProvider<M, N> valueProvider, ListViewAppearance<M> appearance) {
    this.appearance = appearance;
    this.valueProvider = valueProvider;
    setSelectionModel(new ListViewSelectionModel<M>());

    SafeHtmlBuilder markupBuilder = new SafeHtmlBuilder();
    appearance.render(markupBuilder);

    setElement(XDOM.create(markupBuilder.toSafeHtml()));

    getElement().setTabIndex(0);
    getElement().setAttribute("hideFocus", "true");

    all = new CompositeElement();
    setAllowTextSelection(false);

    storeHandlers = new StoreHandlers<M>() {

      @Override
      public void onAdd(StoreAddEvent<M> event) {
        ListView.this.onAdd(event.getItems(), event.getIndex());
      }

      @Override
      public void onClear(StoreClearEvent<M> event) {
        refresh();
      }

      @Override
      public void onDataChange(StoreDataChangeEvent<M> event) {
        refresh();
      }

      @Override
      public void onFilter(StoreFilterEvent<M> event) {
        refresh();
      }

      @Override
      public void onRecordChange(StoreRecordChangeEvent<M> event) {
        if (ListView.this.valueProvider == event.getProperty()) {
          ListView.this.onUpdate(event.getRecord().getModel(), store.indexOf(event.getRecord().getModel()));
        }
      }

      @Override
      public void onRemove(StoreRemoveEvent<M> event) {
        ListView.this.onRemove(event.getItem(), event.getIndex());
      }

      @Override
      public void onSort(StoreSortEvent<M> event) {
        refresh();
      }

      @Override
      public void onUpdate(StoreUpdateEvent<M> event) {
        List<M> items = event.getItems();
        for (int i = 0; i < items.size(); i++) {
          M item = items.get(i);
          ListView.this.onUpdate(item, store.indexOf(item));
        }
      }
    };

    setStore(store);

    sinkEvents(Event.ONCLICK | Event.ONDBLCLICK | Event.MOUSEEVENTS | Event.FOCUSEVENTS | Event.ONKEYDOWN);
  }

  @Override
  public HandlerRegistration addRefreshHandler(RefreshHandler handler) {
    return addHandler(handler, RefreshEvent.getType());
  }

  /**
   * Returns the matching element.
   * 
   * @param element the element or any child element
   * @return the parent element
   */
  public Element findElement(Element element) {
    return appearance.findElement(element.<XElement> cast());
  }

  /**
   * Returns the element's index.
   * 
   * @param element the element or any child element
   * @return the element index or -1 if no match
   */
  public int findElementIndex(Element element) {
    Element elem = findElement(element);
    if (elem != null) {
      return indexOf(elem);
    }
    return -1;
  }

  /**
   * Returns the view's cell.
   * 
   * @return the cell, or null if none specified
   */
  public Cell<N> getCell() {
    return cell;
  }

  /**
   * Returns the element at the given index.
   * 
   * @param index the index
   * @return the element
   */
  public XElement getElement(int index) {
    return all.getElement(index).<XElement> cast();
  }

  /**
   * Returns all of the child elements.
   * 
   * @return the elements
   */
  public List<Element> getElements() {
    return all.getElements();
  }

  /**
   * Returns the number of models in the view.
   * 
   * @return the count
   */
  public int getItemCount() {
    return store == null ? 0 : store.size();
  }

  /**
   * Returns the view's loading text.
   * 
   * @return the loading text
   */
  public String getLoadingText() {
    return loadingText;
  }

  /**
   * Returns the view's quick tip instance.
   * 
   * @return the quicktip instance or null if not enabled
   */
  public QuickTip getQuickTip() {
    return quickTip;
  }

  /**
   * Returns the view's selection model.
   * 
   * @return the selection model
   */
  public ListViewSelectionModel<M> getSelectionModel() {
    return sm;
  }

  /**
   * Returns true if select on hover is enabled.
   * 
   * @return the select on hover state
   */
  public boolean getSelectOnOver() {
    return selectOnHover;
  }

  /**
   * Returns the combo's store.
   * 
   * @return the store
   */
  public ListStore<M> getStore() {
    return store;
  }

  /**
   * Returns the index of the element.
   * 
   * @param element the element
   * @return the index
   */
  public int indexOf(Element element) {
    if (element == null) {
      return -1;
    }
    if (element.getPropertyString("viewIndex") != null) {
      return Util.parseInt(element.getPropertyString("viewIndex"), -1);
    }
    return all.indexOf(element);
  }

  /**
   * Returns true if quicktips are enabled.
   * 
   * @return true for enabled
   */
  public boolean isEnableQuickTips() {
    return enableQuickTip;
  }

  /**
   * Moves the current selections down one level.
   */
  public void moveSelectedDown() {
    List<M> sel = getSelectionModel().getSelectedItems();

    Collections.sort(sel, new Comparator<M>() {
      public int compare(M o1, M o2) {
        return store.indexOf(o1) < store.indexOf(o2) ? 1 : 0;
      }
    });

    for (M m : sel) {
      int idx = store.indexOf(m);
      if (idx < (store.size() - 1)) {
        store.remove(m);
        store.add(idx + 1, m);
      }
    }
    getSelectionModel().select(sel, false);
  }

  /**
   * Moves the current selections up one level.
   */
  public void moveSelectedUp() {
    List<M> sel = getSelectionModel().getSelectedItems();

    Collections.sort(sel, new Comparator<M>() {
      public int compare(M o1, M o2) {
        return store.indexOf(o1) > store.indexOf(o2) ? 1 : 0;
      }
    });

    for (M m : sel) {
      int idx = store.indexOf(m);
      if (idx > 0) {
        store.remove(m);
        store.add(idx - 1, m);
      }
    }
    getSelectionModel().select(sel, false);
  }

  @Override
  public void onBrowserEvent(Event event) {
    if (cell != null) {
      CellWidgetImplHelper.onBrowserEvent(this, event);
    }
    super.onBrowserEvent(event);

    switch (event.getTypeInt()) {
      case Event.ONMOUSEOVER:
        onMouseOver(event);
        break;
      case Event.ONMOUSEOUT:
        onMouseOut(event);
        break;
      case Event.ONMOUSEDOWN:
        onMouseDown(event);
        break;
    }

    handleEventForCell(event);
  }

  /**
   * Refreshes the view by reloading the data from the store and re-rendering
   * the template.
   */
  public void refresh() {
    if (!isOrWasAttached()) {
      return;
    }
    getElement().setInnerHTML("");
    getElement().repaint();

    List<M> models = store == null ? new ArrayList<M>() : store.getAll();
    all.removeAll();

    SafeHtmlBuilder sb = new SafeHtmlBuilder();
    bufferRender(models, sb);

    SafeHtml markup = sb.toSafeHtml();
    getElement().setInnerHTML(markup.asString());

    SafeHtmlBuilder esb = new SafeHtmlBuilder();
    appearance.renderEnd(esb);
    DomHelper.insertHtml("beforeend", getElement(), esb.toSafeHtml().asString());

    List<Element> elems = appearance.findElements(getElement());

    all = new CompositeElement(elems);
    updateIndexes(0, -1);
    fireEvent(new RefreshEvent());
  }

  /**
   * Refreshes an individual node's data from the store.
   * 
   * @param index the items data index in the store
   */
  public void refreshNode(int index) {
    onUpdate(store.get(index), index);
  }

  /**
   * Optionally sets the view's cell. If a cell is not provided, toString is
   * called on the value returned by the view's value provider.
   * 
   * @param cell the cell
   */
  public void setCell(Cell<N> cell) {
    this.cell = cell;

    if (cell != null) {
      CellWidgetImplHelper.sinkEvents(this, cell.getConsumedEvents());
    }
  }

  /**
   * True to enable quicktips (defaults to true, pre-render).
   * 
   * @param enableQuickTip true to enable quicktips
   */
  public void setEnableQuickTips(boolean enableQuickTip) {
    assertPreRender();
    this.enableQuickTip = enableQuickTip;
  }

  /**
   * Sets the text loading text to be displayed during a load request.
   * 
   * @param loadingText the loading text
   */
  public void setLoadingText(String loadingText) {
    this.loadingText = loadingText;
  }

  /**
   * Sets the selection model.
   * 
   * @param sm the selection model
   */
  public void setSelectionModel(ListViewSelectionModel<M> sm) {
    if (this.sm != null) {
      this.sm.bindList(null);
    }
    this.sm = sm;
    if (sm != null) {
      sm.bindList(this);
    }
  }

  /**
   * True to select the item when mousing over a element (defaults to false).
   * 
   * @param selectOnHover true to select on mouse over
   */
  public void setSelectOnOver(boolean selectOnHover) {
    this.selectOnHover = selectOnHover;
  }

  /**
   * Changes the data store bound to this view and refreshes it.
   * 
   * @param store the store to bind this view
   */
  public void setStore(ListStore<M> store) {
    if (this.store != null) {
      handlerRegistration.removeHandler();
    }
    if (store != null) {
      handlerRegistration = store.addStoreHandlers(storeHandlers);
    }
    this.store = store;
    sm.bindList(this);

    if (store != null) {
      refresh();
    }
  }

  protected void bufferRender(List<M> models, SafeHtmlBuilder sb) {
    for (int i = 0, len = models.size(); i < len; i++) {
      M m = models.get(i);
      SafeHtmlBuilder cellBuilder = new SafeHtmlBuilder();
      N v = getValue(m);
      if (cell == null) {
        String text = null;
        if (v != null) {
          text = v.toString();
        }
        cellBuilder.append(Util.isEmptyString(text) ? SafeHtmlUtils.fromTrustedString("&#160;")
            : SafeHtmlUtils.fromString(text));
      } else {
        Context context = new Context(i, 0, store.getKeyProvider().getKey(m));
        cell.render(context, v, cellBuilder);
      }

      appearance.renderItem(sb, cellBuilder.toSafeHtml());
    }
  }

  protected boolean cellConsumesEventType(Cell<N> cell, String eventType) {
    Set<String> consumedEvents = cell.getConsumedEvents();
    return consumedEvents != null && consumedEvents.contains(eventType);
  }

  protected void focusItem(int index) {
    Element elem = getElement(index);
    if (elem != null) {
      elem.<XElement> cast().scrollIntoView(getElement(), false);
    }
    // TODO fix focus issues in IE
    if (!GXT.isIE()) {
      focus();
    }
  }

  protected N getValue(M m) {
    N value;
    if (store.hasRecord(m)) {
      value = store.getRecord(m).getValue(valueProvider);
    } else {
      value = valueProvider.getValue(m);
    }
    return value;
  }

  // @Override
  // public void storeBeforeDataChanged(StoreEvent<M> se) {
  // onBeforeLoad();
  // }

  protected void handleEventForCell(Event event) {
    // Get the event target.
    EventTarget eventTarget = event.getEventTarget();
    if (cell == null || !Element.is(eventTarget)) {
      return;
    }
    final XElement target = event.getEventTarget().cast();

    int rowIndex = findElementIndex(appearance.findElement(target));

    final M m = getStore().get(rowIndex);
    Element cellParent = appearance.findCellParent(getElement(rowIndex));
    if (m != null && cellParent != null) {

      DefaultHandlerManagerContext context = new DefaultHandlerManagerContext(rowIndex, 0,
          getStore().getKeyProvider().getKey(m), ComponentHelper.ensureHandlers(this));

      if (cellConsumesEventType(cell, event.getType())) {
        cell.onBrowserEvent(context, cellParent, getValue(m), event, new ValueUpdater<N>() {

          @Override
          public void update(N value) {
            ListView.this.getStore().getRecord(m).addChange(valueProvider, value);

          }
        });
      }
    }
  }

  protected void onAdd(List<M> models, final int index) {
    if (!isOrWasAttached()) {
      return;
    }

    boolean empty = all.getCount() == 0;
    // add on sorted store, fires multiple adds while store has all models
    // before firing
    if (empty && models.size() == store.size()) {
      refresh();
      return;
    }
    SafeHtmlBuilder sb = new SafeHtmlBuilder();
    bufferRender(models, sb);

    Element d = DOM.createDiv();
    d.setInnerHTML(sb.toSafeHtml().asString());
    List<Element> list = appearance.findElements(d.<XElement> cast());

    final Element ref = index == 0 ? null : all.getElement(index - 1);
    final Element n = ref == null ? null : ref.getParentElement();

    for (int i = list.size() - 1; i >= 0; i--) {
      Element e = list.get(i);
      if (index == 0) {
        getElement().insertFirst(e);
      } else {
        Node next = ref == null ? null : ref.getNextSibling();
        if (next == null) {
          n.appendChild(e);
        } else {
          n.insertBefore(e, next);
        }
      }
    }

    all.insert(Util.toElementArray(list), index);
    updateIndexes(index, -1);
  }

  @Override
  protected void onAfterFirstAttach() {
    super.onAfterFirstAttach();
    if (enableQuickTip) {
      quickTip = new QuickTip(this);
    }
    refresh();
  }

  protected void onBeforeLoad() {
    if (loadingText != null) {
      getElement().setInnerHTML("<div class='loading-indicator'>" + loadingText + "</div>");
      all.removeAll();
    }
  }

  @Override
  protected void onFocus(Event event) {
    super.onFocus(event);
  }

  protected void onHighlightRow(int index, boolean highLight) {
    XElement e = getElement(index);
    if (e != null) {
      e.setClassName("x-view-highlightrow", highLight);
    }
  }

  protected void onMouseDown(Event e) {
    int index = indexOf(e.getEventTarget().<Element> cast());
    if (index != -1) {
      fireEvent(new SelectEvent());
    }
  }

  protected void onMouseOut(Event ce) {
    if (overElement != null) {
      if (!ce.<XEvent> cast().within(overElement, true)) {
        appearance.onOver(overElement, false);
        overElement = null;
      }
    }
  }

  protected void onMouseOver(Event ce) {
    Element target = ce.getEventTarget().<Element> cast();
    target = findElement(target);
    if (target != null) {
      int index = indexOf(target);
      if (index != -1) {
        if (selectOnHover) {
          sm.select(index, false);
        } else {
          if (target != overElement) {
            appearance.onOver(target.<XElement> cast(), true);
            overElement = target.<XElement> cast();
          }
        }
      }
    }
  }

  protected void onRemove(M data, int index) {
    if (all != null) {
      Element e = getElement(index);
      if (e != null) {
        appearance.onOver(e.<XElement> cast(), false);
        if (overElement == e) {
          overElement = null;
        }

        getSelectionModel().deselect(index);

        e.removeFromParent();
        all.remove(index);
        updateIndexes(index, -1);
      }
    }
  }

  protected void onSelectChange(M model, boolean select) {
    if (all != null) {
      int index = store.indexOf(model);
      if (index != -1 && index < all.getCount()) {
        XElement e = getElement(index);
        if (e != null) {
          appearance.onSelect(e, select);
          appearance.onOver(e.<XElement> cast(), false);
        }
      }
    }
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  protected void onUpdate(M model, int index) {
    Element original = all.getElement(index);
    if (original != null) {
      List list = Util.createList(model);
      SafeHtmlBuilder sb = new SafeHtmlBuilder();
      bufferRender(list, sb);

      DomHelper.insertBefore(original, sb.toSafeHtml().asString());

      all.replaceElement(original, Element.as(original.getPreviousSibling()));
      original.removeFromParent();

    }
    sm.onRowUpdated(model);
  }

  protected void updateIndexes(int startIndex, int endIndex) {
    List<Element> elems = all.getElements();
    endIndex = endIndex == -1 ? elems.size() - 1 : endIndex;
    for (int i = startIndex; i <= endIndex; i++) {
      elems.get(i).setPropertyInt("viewIndex", i);
    }
  }

}
