/**
 * Ext GWT 3.0.0-beta1 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.widget.core.client.grid.filters;

import java.util.List;

import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.event.shared.SimpleEventBus;
import com.sencha.gxt.core.client.ValueProvider;
import com.sencha.gxt.data.shared.loader.FilterConfig;
import com.sencha.gxt.data.shared.loader.FilterConfigBean;
import com.sencha.gxt.data.shared.loader.FilterHandler;
import com.sencha.gxt.widget.core.client.event.ActivateEvent;
import com.sencha.gxt.widget.core.client.event.ActivateEvent.ActivateHandler;
import com.sencha.gxt.widget.core.client.event.ActivateEvent.HasActivateHandlers;
import com.sencha.gxt.widget.core.client.event.DeactivateEvent;
import com.sencha.gxt.widget.core.client.event.DeactivateEvent.DeactivateHandler;
import com.sencha.gxt.widget.core.client.event.DeactivateEvent.HasDeactivateHandlers;
import com.sencha.gxt.widget.core.client.event.UpdateEvent;
import com.sencha.gxt.widget.core.client.event.UpdateEvent.HasUpdateHandlers;
import com.sencha.gxt.widget.core.client.event.UpdateEvent.UpdateHandler;
import com.sencha.gxt.widget.core.client.menu.Menu;

public abstract class Filter<M, V> implements HasUpdateHandlers, HasActivateHandlers<Filter<M, ?>>, HasDeactivateHandlers<Filter<M, ?>> {

  protected Menu menu;

  private boolean active = false;
  private SimpleEventBus eventBus;
  private FilterHandler<V> handler;
  private int updateBuffer = 500;
  private final ValueProvider<M, V> valueProvider;
  
  public Filter(ValueProvider<M, V> valueProvider) {
    this.valueProvider = valueProvider;
    
    menu = new Menu();
  }

  @Override
  public HandlerRegistration addActivateHandler(ActivateHandler<Filter<M, ?>> handler) {
    return ensureHandlers().addHandler(ActivateEvent.getType(), handler);
  }
  
  @Override
  public HandlerRegistration addDeactivateHandler(DeactivateHandler<Filter<M, ?>> handler) {
    return ensureHandlers().addHandler(DeactivateEvent.getType(), handler);
  }
  
  @Override
  public HandlerRegistration addUpdateHandler(UpdateHandler handler) {
    return ensureHandlers().addHandler(UpdateEvent.getType(), handler);
  }

  public abstract List<FilterConfig> getFilterConfig();
  
  public FilterHandler<V> getHandler() {
    return handler;
  }

  /**
   * Returns the filter's menu.
   * 
   * @return the menu
   */
  public Menu getMenu() {
    return menu;
  }

  /**
   * Returns the update buffer.
   * 
   * @return the update buffer in milliseconds
   */
  public int getUpdateBuffer() {
    return updateBuffer;
  }

  /**
   * Template method to be implemented by all subclasses that is to get and
   * return the value of the filter.
   */
  public abstract Object getValue();;

  public ValueProvider<M, V> getValueProvider() {
    return valueProvider;
  }

  /**
   * Returns true if the filter is active.
   * 
   * @return the active state
   */
  public boolean isActive() {
    return active;
  }

  /**
   * Sets the status of the filter and fires the appropriate events. You can
   * only set it to active if the filter is activatable.
   * 
   * @param active the new filter state
   * @param supressEvent true to prevent events from being fired
   */
  public void setActive(boolean active, boolean supressEvent) {
    active = active && isActivatable();
    if (this.active != active) {
      this.active = active;
      if (!supressEvent) {
        if (active) {
          ensureHandlers().fireEventFromSource(new ActivateEvent<Filter<M, ?>>(this), this);
        } else {
          ensureHandlers().fireEventFromSource(new DeactivateEvent<Filter<M, ?>>(this), this);
        }
      }
    }
  }

  public void setHandler(FilterHandler<V> handler) {
    this.handler = handler;
  }
  
  /**
   * Number of milliseconds to wait after user interaction to fire an update
   * (defaults to 500).
   * 
   * @param updateBuffer the update buffer in milliseconds
   */
  public void setUpdateBuffer(int updateBuffer) {
    this.updateBuffer = updateBuffer;
  }

  protected FilterConfig createNewFilterConfig() {
    FilterConfigBean cfg = new FilterConfigBean();
    cfg.setFieldAndType(getValueProvider(), getType());
    return cfg;
  }

  protected void fireUpdate() {
    if (active) {
      ensureHandlers().fireEventFromSource(new UpdateEvent(), this);
    }
    setActive(isActivatable(), false);
  }

  protected abstract Class<V> getType();

  /**
   * Template method to be implemented by all subclasses that is to return
   * <code>true</code> if the filter has enough configuration information to be
   * activated.
   * 
   * @return true if if the filter has enough configuration information to be
   *         activated
   */
  protected boolean isActivatable() {
    return true;
  }
  
  /**
   * Template method to be implemented by all subclasses that is to validates
   * the provided Model against the filters configuration. Defaults to
   * <tt>return true</tt>.
   * 
   * @param model the model
   * @return true if valid
   */
  protected boolean validateModel(M model) {
    return true;
  }
  
  SimpleEventBus ensureHandlers() {
    return eventBus == null ? eventBus = new SimpleEventBus() : eventBus;
  }
}
